# Gcode_importer.rb - Imports gcode .gcode file in SketchUp
#
# Copyright (C) 2013 David Fishelzon
#
# License: Apache License, Version 2.0
#
# G-Code Importer Version 0.0.3
#
require 'sketchup'
Sketchup.send_action "showRubyPanel:"

module GcodeImporterModule

  class GcodeImporter < Sketchup::Importer

    # Find the "filament" component if it exist OR
    # if the make a "filament" component if it does not exist
    def Get_Filament_component
      @filamentComp = Sketchup.active_model.definitions[ "filament" ]
      unless @filamentComp
        # if there is no existing component named "filament", make one
        # We will make one
        @nozel        = 0.4.mm
        @layerH       = 1.0.mm   # the Z axis movements will scale this parameter
        @filamentComp = Sketchup.active_model.definitions.add("filament")
        @pts          = [  ORIGIN,
                 [-@nozel/2.0, 0.0.mm , -@layerH/2.0], 
                 [-@nozel/2.0, 0.0.mm , -@layerH    ], 
                 [ @nozel/2.0, 0.0.mm , -@layerH    ], 
                 [ @nozel/2.0, 0.0.mm , -@layerH/2.0] ]
        @newface = @filamentComp.entities.add_face(@pts)
        @newface.casts_shadows    = false
        @newface.receives_shadows = false
        @newface.reverse! if @newface.normal.y < 0
        @newface.pushpull(1.0.mm)
        @filamentComp.entities.each { |@ent|
          if @ent.type == Sketchup::Edge
            @ent.hidden        = true
            @ent.casts_shadows = false
          end # if edge
        } # each ent
      end # unless @filamentComp

      @HeadMoveComp = Sketchup.active_model.definitions[ "HeadMove" ]
      unless @HeadMoveComp
        @HeadMoveComp = Sketchup.active_model.definitions.add("HeadMove")
        @oldPoint = [0.0, 0.0 , 0.0]
        @newPoint = [0.0, 1.mm, 0.0]
        @HeadMoveComp.entities.add_line @oldPoint, @newPoint
      end # unless @HeadMoveComp
    end # def Get_Filament_component

    # Adds a "filament" component between @oldPoint and @newPoint
    def MoveOld2newPoint( comp )
      @lineVector  = @oldPoint.vector_to @newPoint
      return nil if @lineVector.length.to_f < 1e-6 

      # Scaling and adding an instance of the filament
      @scaleL        = @lineVector.length.to_f / @filamentComp.bounds.height   #width #depth 
      return nil if @scaleL < 1e-6 or @scaleZ < 1e-6
      @trans         = Geom::Transformation.scaling 1.0, @scaleL, @scaleZ
      @instance      = Sketchup.active_model.active_entities.add_instance( comp, @trans )

      # if needed rotate the filament
      @lineVector.normalize!
      @instYaxis     = Geom::Vector3d.new( 0.0, @filamentComp.bounds.height, 0.0 ).normalize #height givs the Y axis
      @rotDeg        = @instYaxis.dot( @lineVector )
      @rotDeg        = Math::acos( @rotDeg )
      if @rotDeg.to_f - 1e-6 > 0.0
        if @rotDeg.to_f + 1e-6 > Math::PI 
          @rotVector = Geom::Vector3d.new( 0.0, 0.0, 1.0 )
        else # if PI
          @rotVector = @instYaxis.cross( @lineVector )
        end # if PI
        @trans       = Geom::Transformation.rotation( ORIGIN, @rotVector, @rotDeg )
        @instance.transform!( @trans )
      end # if @rotDeg > 0.0

      # Moving the filament to the @oldPoint
      if not ( ORIGIN == @oldPoint )
        @trans       = Geom::Transformation.new @oldPoint 
        @instance.transform!( @trans )
      end #if not ORIGEN

      @instance.casts_shadows    = false
      @instance.receives_shadows = false
      return @instance
    end # def move_old2new

    def initialize
      Get_Filament_component()      # Add a filament component
    end

    def description
      "Gcode Importer (*.gcode;*.bfb) by David Fishelzon"
    end

    def id
      "DavidFishelzon@Yahoo.com/Gcode_importer"
    end

    def file_extension
      return "*.gcode;*.bfb"
    end

    def supports_options?
      false
    end

    def do_options
      false
    end

    def load_file( filename, status )
      if status == true
        begin
          if (Sketchup.version_number < 7)
            Sketchup.active_model.start_operation("Import G-code")
          else
            Sketchup.active_model.start_operation("Import G-code",true)
          end
          Sketchup.active_model.active_view.dynamic = true
          gcode_import(filename)
          Sketchup.active_model.commit_operation # commits an operation for undo
          puts "FINISH FINISH FINISH FINISH"
          Sketchup.status_text=""
          return 0 # Success
        rescue Exception=>e
          puts "!!!!!!! !!!!!!!!!!!!!!!!!!!!!!!!! !!!!!!!"
          puts "!!!!!!! Bug in the Gcode importer !!!!!!!"
          puts "!!!!!!! !!!!!!!!!!!!!!!!!!!!!!!!! !!!!!!!"
          puts e.message
          puts e.backtrace
          # wait forever
          while 1==1 do 
          end # while 1
        end # begin rescue
      end # if status
    end # def load_file

    def gcode_import( filename )
      @model       = Sketchup.active_model
      @entities    = @model.entities
      @x = @y = @z = @f = 0.0       # x,y,z are positions f is the speed of movement
      @scaleZold   = @scaleZ = 0.4  # Default Layer Hight is 0.4mm
      @units       = 1.0.mm         # Default units is mm
      @e           = 1.0 # extruder on by default 
      @oldE        = 1.0 # in absolute length we need to get only the extruded length
      @incremental = false
      @extruderIncremental     = true
      @oldPoint    = @newPoint = [ @x, @y, @z ]
      @lineNumber  = 0

      Sketchup.status_text=" Starting G Code Import "
      IO.foreach(filename) do |@line|
        @lineNumber += 1
        if @lineNumber % 1000 == 0
          Sketchup.status_text=" #{@lineNumber} Lines imported "
          Sketchup.active_model.active_view.zoom_extents if @lineNumber == 10000 
          Sketchup.active_model.active_view.refresh      if @lineNumber == 10000 
        end #if
 
        @line = @line[/[^;]*/].strip   # remuve the ";" remarks from the line
        if      @line.length < 2       # not enuf chars in this line 
          next                         # move to next line
        end # if

        # execute the Command:
        @command, *@params = @line.split
        case @command

        when 'G1', 'G92', 'G28', 'M92' # G1 move motors: G92 sets an origin offset
          # read the line by adding, befor the first number(/\d/), an '=' and evaluating 
          @params.map {|@var| eval('@' + @var.insert(@var.index(/[-\d]/),'=').downcase + '*@units' ) if @var[/[xyzefXYZEF]/] }

          if @incremental == true
            @newPoint = [@x+@oldPoint[0],@y+@oldPoint[1],@z+@oldPoint[2]]
          else # ;absolute positioning
            @newPoint = [@x             ,@y             ,@z             ]
          end  # if @incremental 

          if @command == 'G92'
            @oldE     = @e
          else #@command == 'G92'
            if @extruderIncremental == true
              # e is the extruded length
            else # if @extruderIncremental 
              # absolute extrution
              @e      = @e - @oldE # calculate extruded length
              @oldE   = @e + @oldE
            end # if @extruderIncremental 
          end  #@command == 'G92'

          if not (@z - @oldPoint[2]).between?( -1e-6, 1e-6)
            # movement in the Z direction - update the scaleZ
            if Geom::Vector3d.new( @x - @oldPoint[0], @y - @oldPoint[1], 0.0 ).length.to_f < 10.mm
            # only move in Z
              @scaleZ      = (@z - @oldPoint[2]) / @filamentComp.bounds.depth
              @scaleZ      = -@scaleZ   if @scaleZ < 0.0 
              @scaleZ      = @scaleZold if @scaleZ > 1.0 # if the new scale is not reasnable - don't use it
              @scaleZold   = @scaleZ                     # remember the defoult high
            end # if x,y moved
          end  # if 

          if @e <= 0.0
            # No extruded filament - so just draw a line
            @lastMove    = MoveOld2newPoint( @HeadMoveComp )
          else # @e <= 0.0
            # Extruded filament 
            # don't move in the Z axis
            @saveZ       = @newPoint[2]
            @newPoint[2] = @oldPoint[2]
            @lastMove    = MoveOld2newPoint( @filamentComp )
            @newPoint[2] = @saveZ
          end  # @e <= 0.0

          if @lastMove
            @oldPoint     = @newPoint
            @lastMove.name= " #{@lineNumber} " + @line
            #@lastMove.set_attribute "physical_properties", "Motor_Move", "MoveHead"
          else # if lastMove
            if not Geom::Vector3d.new(@newPoint[0] - @oldPoint[0],
                                      @newPoint[1] - @oldPoint[1],
                                      @newPoint[2] - @oldPoint[2]
                                     ).length.to_f.between?( -1e-6, 1e-6 )
              puts "#{@lineNumber} Unable to add line: >#{@line}< e=#{@e} "
              puts "x=#{@oldPoint[0]/@units} y=#{@oldPoint[1]/@units} z=#{@oldPoint[2]/@units} "
            end # if @line[/[xyzXYZ]/]
          end  # if lastMove
        ############################# end of when 'G1', 'G92', 'G28', 'M92' 
 
        when 'M103' # extruder off
          @e = 0.0 

        when 'M101' # extruder on 
          @e = 1.0

        when 'M82'
          #: set extruder to absolute mode 
          @extruderIncremental = false

        when 'M83'
          #: set extruder to relative mode 
          @extruderIncremental = true

        when 'G90' # G90 Absolute Positioning Implemented.
          @incremental = false

        when 'G91' # Incremental mode
          @incremental = true

        when 'G20'
          @units = 1.0.inch

        when 'G21'
          @units = 1.0.mm

        when 'M106', 'M107'
          #M106 Turn fan on.
          #M107 Turn fan off.
        when 'M104'
          #M104 S247.0 Set temperature to 145.0 DegC
        when 'M84'
          #Stop the idle hold on all axis and extruder - Noise reduction
        when 'M108'
          # S400 Set Extruder 1 speed to S value/10 = 40rpm.
        when 'M109'
          #: Set Extruder Temperature and Wait

        else 
          # disply a line what we did not understand
          puts "#{@lineNumber} New type of command: #{@line} "

        end # of case command
      end # of do |line|
      Sketchup.status_text="Done..."
    end # of def gcode_import( filename )

  end # class GcodeImporter < Sketchup::Importer

  Sketchup.register_importer(GcodeImporter.new)
end # GcodeImporterModule


